`timescale 1 ps / 1 ps

// SNS Event Link client simulation.
//  module fpga    is expected to be synthesized
//  module ad9901  represents the external phase comparision chip
//  module main    is there for test bench purposes
// March, 2001   Larry Doolittle   <LRDoolittle@lbl.gov>

// The test bench part of this simulation refers to PLI functions
// $filter and $period, in which I have modelled the filter/VCO pair.
// Source code is in filter.c, loadable module filter.vpi.

// Symbol decode and phase locking is conceptually simple.
// Adding provision for frequency locking made this version,
// at least, a lot uglier.  That's what advance and retard
// are for.
module fpga(clk, inp, phase2, phcomp);
   input clk, inp;
   output phase2, phcomp;
   wire phcomp, advance, retard;
   reg phase1, phase2, d0, d1, d2, d3, retardl, advancel;
   reg prev, data, error;
   initial phase1=0;
   initial phase2=0;
   initial error=0;
   initial retardl=0;
   initial advancel=0;
   initial d0=0;
   initial d1=0;
   initial d2=0;
   initial d3=0;
   assign advance = d1==d3 & d1 != d2 & phase1 & phase2;
   assign retard  = d1==d3 & d1 == d2 & phase1 & phase2 & ~retardl;
   always @(posedge clk) begin
      phase1 <= ~phase2 | (phase1 & phase2 & retard);
      phase2 <= phase1 & ~advance;
      if (~phase1 & ~phase2) d0 <= inp;
      if ( phase1 & ~phase2) d1 <= inp;  // old qlatch
      if ( phase1 &  phase2 & ~retard) d2 <= inp;
      if ((~phase1 &  phase2) | advance) d3 <= inp;  // old ilatch

// If at time 2 we notice that d1 == d3, and they are both
// equal to the old d2, we should delay our clocks by a phase
// (such as have two phase 2's).
// But if d1 == d3, and they are both _not_ equal to the old d2,
// we should advance our clocks a phase (e.g., skip directly to time 3). 
      if ( phase1 &  phase2) retardl <= retard;
      advancel <= advance;

      if (~phase1 & ~phase2) prev  <= d1;
      if (~phase1 & ~phase2) data  <= d1 ^ prev;
      if ( phase1 &  phase2) error <= ~( d3 ^ d1 );
   end
   assign phcomp = d3 ^ inp ^ advancel;
endmodule

// The two inputs to an AD9901 are called "reference" and "oscillator",
// hence all the r and o suffixes.
module ad9901(in_r, in_o, ot);
   input in_r, in_o;
   output ot;
   reg st_r, st_o;   // input flip-flops
   reg ds_r, ds_o;   // frequency discriminator flip-flops
   initial st_r=0;
   initial st_o=0;
   initial ds_r=0;
   initial ds_o=1;
   wire xr, ot;
   assign xr = ~(st_r ^ st_o);
   always @(posedge in_r) st_r <= ~st_r;
   always @(posedge in_o) st_o <= ~st_o;

   // in the locked state, we expect
   //   posedge in_r to happen when xr is 0 (and becomes 1), and
   //   posedge in_o to happen when xr is 1 (and becomes 0).
   // ~ds_o is the asynchronous reset input for ds_r
   always @(posedge in_r or negedge ds_o) ds_r <= xr & ds_o;
   //  ds_r is the asynchronous set input for ds_o
   always @(posedge in_o or posedge ds_r) ds_o <= xr | ds_r;
   assign ot = (xr & ds_o) | ds_r ;
endmodule

module main();
   reg vco, inp;
   wire phase1, phcomp, phdiff;
   integer cc;   // cycle count
   integer high_time;
   initial begin
      $dumpfile("loop.vcd");
      $dumpvars(3,main);
   end
   // VCO period computed from PLI module
   always begin
      high_time=$vco_period($time)-7000;
      vco=0;
      #high_time;
      vco=1;
      #7000;
   end
   // Model the input as a random stream of marks and spaces.
   // Take time seriously: 945ns is 16 bit times,
   // so each half of the cycle is nominally 29531 ps.
   // The extreme values are:
   //     70.53 MHz -> 28356 ps
   //     65.34 MHz -> 30609 ps
   reg data_bit;
   initial begin
      inp = 0;
      #100
      for (cc=0; cc<2000; cc=cc+1) begin
         inp<=~inp;
         #29531;
         data_bit <= $random<2147483648;
         if (cc>1000 && ~data_bit) inp<=~inp;
         // inp<=~inp;
         #29531;
         if ((cc %4) == 1) #1;
      end
      $finish;
   end

   reg q1,q2;
   initial q1=0;
   initial q2=0;
   always @(posedge vco) begin
      q1<=~q2;
      q2<=q1;
   end
   fpga u1(vco, inp, phase1, phcomp);
   ad9901 u2(q1, phcomp, phdiff);
   // ad9901 u2(phase1, phcomp, phdiff);
   // ad9901 u2(phase1, inp, phdiff);
   // always @(phdiff) $display("%d ns  %d",$time,phdiff);
   always @(phdiff) $filter($time,phdiff);
endmodule
